/*
 * proxy.c:
 * Proxy operation for tlsproxyd.
 *
 * Copyright (c) 2002 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: proxy.c,v 1.6 2002/07/05 14:22:44 chris Exp $";

#include <sys/types.h>

#include <fcntl.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "tlsproxyd.h"

sig_atomic_t nprocs, foad, restart;

jmp_buf timeout_abort;

/* sigchld_handler:
 * Handler for SIGCHLD. */
static void sigchld_handler(const int s) {
    pid_t pid;
    int e, status;
    e = errno;  /* save errno */

    while (1) {
        pid = waitpid(-1, &status, WNOHANG);
        if (pid > 0)
            --nprocs;
        else {
            errno = e;
            return;
        }
    }
}

/* sigalrm_handler:
 * Handler for SIGALRM: nonlocal jump to timeout handler. */
static void sigalrm_handler(const int s) {
    siglongjmp(timeout_abort, 1);
}

/* sig_terminate_handler:
 * Handler for signals which cause the daemon to terminate. */
static void sig_terminate_handler(const int s) {
    foad = s;
}

/* sig_restart_handler:
 * Handler for signals which cause the daemon to restart. */
static void sig_restart_handler(const int s) {
    restart = s;
}

/* connection_str:
 * Return a string describing a connection. */
static char *connection_str(int s_from, const struct sockaddr_in *a_to) {
    static char buf[512] = "", *p;
    struct sockaddr_in sin;
    socklen_t n;

    p = buf;
    n = sizeof sin;
    getpeername(s_from, &sin, &n);
    sprintf(p, "%s:%d -", inet_ntoa(sin.sin_addr), (int)ntohs(sin.sin_port));     p += strlen(p);
    n = sizeof sin;
    getsockname(s_from, &sin, &n);
    sprintf(p, "/%s:%d/-> ", inet_ntoa(sin.sin_addr), (int)ntohs(sin.sin_port));   p += strlen(p);

    sprintf(p, "%s:%d", inet_ntoa(a_to->sin_addr), (int)ntohs(a_to->sin_port));

    return buf;
}

/* ssl_error_string:
 * Get a useful error string from OpenSSL. */
const char *ssl_error_string(SSL *ssl, int r) {
    int err;
    err = SSL_get_error(ssl, r);
    if (err == SSL_ERROR_SSL)
        return ERR_reason_error_string(ERR_get_error());
    else
        return ERR_reason_error_string(err);
}


/* do_proxy:
 * Run the proxy loop in an individual child process.
 * XXX close/timeout handling is flaky here, but probably good enough for our
 * purposes. */
#define BUFSIZE  1024
int do_proxy(int s_from, const struct sockaddr_in *a_to, struct config *cfg, SSL_CTX *ctx) {
    int s_to;
    struct proxyent *pe;
    SSL *ssl;
    int nn;
    int r, c2sl = 0, c2s_offset = 0;
    int read_blocked_on_write = 0, write_blocked_on_read = 0, read_blocked = 0;
    int shutdown_wait = 0;
    char c2s[BUFSIZE], s2c[BUFSIZE];
    int closed_err = 0;
    fd_set rdfds, wrfds;

    xsignal(SIGCHLD, SIG_IGN);
    
    /* Close listening sockets. */
    for (pe = cfg->proxy; pe < cfg->proxy + cfg->nproxy; ++pe)
        close(pe->s);
 
    /* Timeout abort operation. */
    xsignal(SIGALRM, sigalrm_handler);
    if (sigsetjmp(timeout_abort, 0)) {
        log_print(LOG_ERR, "%s: timed out reading from server", connection_str(s_from, a_to));
        closed_err = 1;
        goto end;
    }

    alarm(cfg->timeout);

    /* Start up SSL on the client connection. */
    ssl = SSL_new(ctx);
    if (!ssl) {
        log_print(LOG_ERR, "%s: SSL_new: %s", connection_str(s_from, a_to), ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free(ctx);
        _exit(1);
    }

    SSL_set_fd(ssl, s_from);
    fcntl(s_from, F_SETFL, 0);

    /* Do SSL negotiation. */
    if ((r = SSL_accept(ssl)) <= 0) {
        log_print(LOG_ERR, "%s: SSL_accept: %s", connection_str(s_from, a_to), ssl_error_string(ssl, r));
        SSL_free(ssl);
        SSL_CTX_free(ctx);
        close(s_to);
        close(s_from);
        _exit(1);
    }

    fcntl(s_from, F_SETFL, O_NONBLOCK);
    
    alarm(cfg->timeout);

    /* SSL successfully negotiated, so connect to the non-secured service. */
    s_to = socket(PF_INET, SOCK_STREAM, 0);
    if (s_to == -1) {
        log_print(LOG_ERR, "%s: socket: %m", connection_str(s_from, a_to));
        close(s_from);
        _exit(1);
    } else if (connect(s_to, (struct sockaddr*)a_to, sizeof *a_to) == -1) {
        log_print(LOG_ERR, "%s: connect: %m", connection_str(s_from, a_to));
        shutdown(s_from, 1);
        close(s_from);
        close(s_to);
        _exit(1);
    } else
        log_print(LOG_INFO, "%s: established new connection", connection_str(s_from, a_to));

    /* Actually run the proxy loop; this is based upon the code from Eric
     * Rescorla's OpenSSL tutorial, at http://www.rtfm.com/openssl-examples/ */
    nn = s_from > s_to ? s_from : s_to;
    ++nn;

    while (!foad) {
        struct timeval tv;

        tv.tv_sec = 1;
        tv.tv_usec = 0;
        
        /* Change timeout handling here. */
        if (sigsetjmp(timeout_abort, 0)) {
            log_print(LOG_ERR, "%s: timed out reading from server", connection_str(s_from, a_to));
            shutdown(s_to, 2);
            close(s_to);
            s_to = -1;
            if (!shutdown_wait)
                SSL_shutdown(ssl);
            closed_err = 1;
            goto end;
        }
        
        FD_ZERO(&rdfds);
        FD_ZERO(&wrfds);

        FD_SET(s_from, &rdfds);

        /* If we're waiting for a read on the socket don't try to write to the
         * server. */
        if (!write_blocked_on_read) {
            /* If we have data in the write queue don't try to read from peer. */
            if (c2sl || read_blocked_on_write)
                FD_SET(s_from, &wrfds);
            else if (s_to != -1)
                FD_SET(s_to, &rdfds);
        }

        r = select(nn, &rdfds, &wrfds, NULL, &tv);
        if (r == 0)
            continue;
        else if (r == -1 && errno != EAGAIN) {
            log_print(LOG_ERR, "%s: select: %m; aborting", connection_str(s_from, a_to));
            closed_err = 1;
            goto end;
        }

        /* Now check if there's data to read */
        if ((FD_ISSET(s_from, &rdfds) && !write_blocked_on_read) || (read_blocked_on_write && FD_ISSET(s_from, &wrfds))){
            do {
                read_blocked_on_write = 0;
                read_blocked = 0;

                r = SSL_read(ssl, s2c, BUFSIZE);

                switch (SSL_get_error(ssl, r)){
                    case SSL_ERROR_NONE:
                        /* Note: this call could block, which blocks the
                         * entire application. This is fine, though. */
                        write(s_to, s2c, r);
                        break;
                        
                    case SSL_ERROR_ZERO_RETURN:
                        /* End of data: peer closed the connection. */
                        if (!shutdown_wait)
                            SSL_shutdown(ssl);
                        shutdown(s_to, 2);
                        close(s_to);
                        goto end;
                        
                    case SSL_ERROR_WANT_READ:
                        read_blocked = 1;
                        break;

                    case SSL_ERROR_WANT_WRITE:
                        /* We get a WANT_WRITE if we're trying to rehandshake
                         * and we block on a write during that rehandshake.
                         * We need to wait on the socket to be writeable but
                         * reinitiate the read when it is */
                        read_blocked_on_write = 1;
                        break;

                    default:
                        log_print(LOG_ERR, "%s: SSL_read: %s", connection_str(s_from, a_to), ssl_error_string(ssl, r));
                        closed_err = 1;
                        goto end;
                }

                /* We need a check for read_blocked here because SSL_pending()
                 * doesn't work properly during the handshake. This check
                 * prevents a busy-wait loop around SSL_read() */
            } while (SSL_pending(ssl) && !read_blocked);
        }

        /* Check for data from the server. */
        if (FD_ISSET(s_to, &rdfds)) {
            c2sl = read(s_to, c2s, BUFSIZE);
            /* Server closed connection. */
            if (c2sl == -1) {
                log_print(LOG_ERR, "%s: read from server: %m; closing connection", connection_str(s_from, a_to));
                closed_err = 1;
            } else if (c2sl == 0) {
                close(s_to);
                s_to = -1;
                shutdown_wait = 1;
                if (SSL_shutdown(ssl))
                    break;
            }
            c2s_offset = 0;
        }

        /* If the socket is writeable.... */
        if ((FD_ISSET(s_from, &wrfds) && c2sl) || (write_blocked_on_read && FD_ISSET(s_from, &rdfds))) {
            write_blocked_on_read = 0;

            /* Try to write. */
            r = SSL_write(ssl, c2s + c2s_offset, c2sl);

            switch(SSL_get_error(ssl,r)){
                case SSL_ERROR_NONE:
                    /* We wrote something. */
                    c2sl -= r;
                    c2s_offset += r;
                    break;

                case SSL_ERROR_WANT_WRITE:
                    /* We would have blocked. */
                    break;

                case SSL_ERROR_WANT_READ:
                    /* We get a WANT_READ if we're trying to rehandshake and
                     * we block on write during the current connection. We need
                     * to wait on the socket to be readable but reinitiate our
                     * write when it is. */
                    write_blocked_on_read=1;
                    break;
                    
                default:
                    log_print(LOG_ERR, "%s: SSL_read: %s", connection_str(s_from, a_to), ERR_reason_error_string(ERR_get_error()));
                    goto end;
            }
        }
    }

end:
    if (closed_err)
        log_print(LOG_ERR, "%s: connection shut down after error", connection_str(s_from, a_to));
    else
        log_print(LOG_INFO, "%s: connection shut down normally", connection_str(s_from, a_to));
    
    SSL_free(ssl);
    close(s_from);
    close(s_to);

    _exit(0);
}

/* do_proxy_accept:
 * Accept connections, forking child processes to handle them. */
int do_proxy_accept(struct config *cfg, SSL_CTX *ctx) {
    int term_signals[] = {SIGTERM, SIGINT, 0};
    int restart_signals[] = {SIGHUP, 0};
    int ignore_signals[] = {SIGUSR1, SIGUSR2, SIGPIPE, SIGALRM, SIGIO, 0};
    int *p;
        
    xsignal(SIGCHLD, sigchld_handler);
    for (p = term_signals; *p; ++p)
        xsignal(*p, sig_terminate_handler);
    for (p = restart_signals; *p; ++p)
        xsignal(*p, sig_restart_handler);
    for (p = ignore_signals; *p; ++p)
        xsignal(*p, SIG_IGN);
    
    /* Log some info. */
    log_print(LOG_INFO, "started tlsproxyd, listening on %d ports", cfg->nproxy);
    
    while (!foad && !restart) {
        int n = 0;
        struct timeval tv;
        struct proxyent *pe;
        fd_set rdfds;

        tv.tv_sec = 1;
        tv.tv_usec = 0;

        if (nprocs < cfg->maxprocs) {
            for (pe = cfg->proxy; pe < cfg->proxy + cfg->nproxy; ++pe) {
                FD_SET(pe->s, &rdfds);
                if (pe->s >= n) n = pe->s + 1;
            }
        }

        switch (select(n, &rdfds, NULL, NULL, &tv)) {
            case -1:
                /* Error. */
                if (errno != EINTR) {
                    log_print(LOG_ERR, "select: %m; aborting");
                    _exit(1);
                }
                break;

            case 0:
                break;

            default:
                block_signal(SIGCHLD);

                /* Accept connections and fork children to handle them. */
                for (pe = cfg->proxy; pe < cfg->proxy + cfg->nproxy; ++pe) {
                    if (FD_ISSET(pe->s, &rdfds)) {
                        int s_from;
                        struct sockaddr_in sin;
                        socklen_t n;

                        n = sizeof sin;
                        s_from = accept(pe->s, &sin, &n);
                        fcntl(s_from, F_SETFL, O_NONBLOCK);
                        if (s_from == -1 && errno != EAGAIN) {
                            log_print(LOG_ERR, "%s:%d: accept: %m", inet_ntoa(pe->from.sin_addr), (int)ntohs(pe->from.sin_port));
                            continue;
                        } else if (nprocs >= cfg->maxprocs) {
                            log_print(LOG_ERR, "%s:%d: too busy; closing connection", inet_ntoa(pe->from.sin_addr), (int)ntohs(pe->from.sin_port));
                            close(s_from);
                        } else {
                            switch (fork()) {
                                case 0:
                                    /* Child. */
                                    log_print(LOG_ERR, "%s: new incoming connection", connection_str(s_from, &pe->to));
                                    do_proxy(s_from, &pe->to, cfg, ctx);
                                    _exit(0);
                                    break;  /* NOTREACHED */
                                    
                                case -1:
                                    log_print(LOG_ERR, "fork: %m; aborting");
                                    _exit(1);
                                    break;  /* NOTREACHED */

                                default:
                                    close(s_from);
                                    ++nprocs;
                                    break;
                            }
                        }
                    }
                }
                unblock_signal(SIGCHLD);
        }
    }

    if (foad) {
        log_print(LOG_INFO, "terminated by signal %d", (int)foad);
        return 0;
    } else if (restart) {
        log_print(LOG_INFO, "restarting on signal %d", (int)restart);
        return 1;
    }

    return 0; /* NOTREACHED */
}

