/*
 * tlsproxyd.c:
 * TLS proxy.
 *
 * Copyright (c) 2002 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: tlsproxyd.c,v 1.5 2002/06/11 13:52:19 chris Exp $";

#include <fcntl.h>
#include <getopt.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>

#include "tlsproxyd.h"

static int noreadpassphrase;

/* tls_getpass:
 * Callback function to get a pass phrase from the user. */
int tls_getpass(char *buf, int size, int rwflag, void *userdata) {
    char *s;
    memset(buf, 0, size);
    if (noreadpassphrase) return 0;
    fprintf(stderr, "%s\n", (char*)userdata);
    s = getpass("Enter pass phrase: ");
    strncpy(buf, s, size - 1);
    return strlen(buf);
}

/* tls_start:
 * Start up the TLS stuff. */
SSL_CTX *tls_start(struct config *cfg) {
    SSL_CTX *ctx;
    int ret;
    
    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();

    /* Create context supporting all protocols. */
    ctx = SSL_CTX_new(SSLv23_server_method());
    if (!ctx) {
        log_print(LOG_ERR, "unable to create SSL context");
        return NULL;
    }

    /* Set up the password call back. */
    SSL_CTX_set_default_passwd_cb(ctx, tls_getpass);
    SSL_CTX_set_default_passwd_cb_userdata(ctx, (void*)cfg->certfile);

    /* Load certificate, and, if specified, separate private key. */
    if ((ret = SSL_CTX_use_certificate_file(ctx, cfg->certfile, SSL_FILETYPE_PEM)) <= 0) {
        log_print(LOG_ERR, "%s: %s", cfg->certfile, ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free(ctx);
        return NULL;
    }
    
    SSL_CTX_set_default_passwd_cb_userdata(ctx, (void*)cfg->pkeyfile);
    if ((ret = SSL_CTX_use_PrivateKey_file(ctx, cfg->pkeyfile, SSL_FILETYPE_PEM)) <= 0) {
        log_print(LOG_ERR, "%s: %s", cfg->pkeyfile, ERR_reason_error_string(ERR_get_error()));
        SSL_CTX_free(ctx);
        return NULL;
    }

    /* Verify that the private key matches the certificate. */
    if (!SSL_CTX_check_private_key(ctx)) {
        log_print(LOG_ERR, "private key does not match certificate public key");
        SSL_CTX_free(ctx);
        return NULL;
    }

    return ctx;
}

/* pidfile_create:
 * Create and lock a PID file. Returns 1 on success or 0 on failure. */
int fd_pidfile = -1;

int pidfile_create(const struct config *cfg) {
    fd_pidfile = open(cfg->pidfile, O_RDWR | O_CREAT, 0644);
    if (fd_pidfile == -1) {
        log_print(LOG_ERR, "%s: could not create PID file: %m", cfg->pidfile);
        return 0;
    } else if (fcntl_lock(fd_pidfile) == -1) {
        log_print(LOG_ERR, "%s: could not lock PID file: %m; probably another process is using it", cfg->pidfile);
        return 0;
    } else {
        /* OK. */
        ftruncate(fd_pidfile, 0);
        return 1;
    }
}

/* pidfile_write_pid:
 * Write PID into PID file. */
int pidfile_write_pid(void) {
    char buf[16];
    if (fd_pidfile == -1)
        return 0;
    sprintf(buf, "%d\n", (int)getpid());
    return (xwrite(fd_pidfile, buf, strlen(buf)) == strlen(buf));
}

/* pidfile_remove:
 * Remove PID file. */
int pidfile_remove(const struct config *cfg) {
    return unlink(cfg->pidfile);
}

/* usage:
 * Print usage information. */
void usage(FILE *fp) {
    fprintf(fp,
"tlsproxyd: TLS proxy service\n"
"\n"
"Synopsis: tlsproxyd -h | [-d] [-f file] [-p]\n"
"\n"
"    -h       display this message\n"
"    -d       do not detach from controlling terminal\n"
"    -f file  read configuration from file\n"
"             default: "CFGFILE"\n"
"    -p       do not attempt to read a pass phrase for encrypted keys\n"
"\n"
"Copyright (c) 2002 Chris Lightfoot <chris@ex-parrot.com>\n"
"http://www.ex-parrot.com/~chris/tlsproxyd/\n"
"\n"
"This program is free software; you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation; either version 2 of the License, or\n"
"(at your option) any later version.\n"
"\n"
            );
}

/* main:
 * Entry point. */
char optstr[] = "dhf:p";
extern int log_stderr, log_syslog; /* in util.c */

int main(int argc, char *argv[]) {
    char *cfgfile = CFGFILE;
    struct config *cfg;
    int c, do_daemon = 1;
    struct proxyent *pe;
    int restart;
    SSL_CTX *ctx;

    opterr = 0;
    while ((c = getopt(argc, argv, optstr)) != EOF) {
        switch (c) {
            case 'h':
                usage(stdout);
                return 0;

            case 'd':
                do_daemon = 0;
                break;

            case 'f':
                cfgfile = optarg;
                break;

            case 'p':
                noreadpassphrase = 1;
                break;

            case '?':
            default:
                if (strchr(optstr, optopt))
                    fprintf(stderr, PROGNAME": option -%c requires an argument\n", optopt);
                else
                    fprintf(stderr, PROGNAME": unrecognised option -%c\n", optopt);
                usage(stderr);
                return 1;
        }
    }

    if (!isatty(0))
        noreadpassphrase = 1;


    if (!(cfg = read_config(cfgfile))) {
        log_print(LOG_ERR, "configuration errors; aborting");
        return -1;
    }

    /* Start up logging. */
    openlog("tlsproxyd", LOG_PID, cfg->logfacil);
    log_syslog = 1;

    /* Set up network stuff. */
    for (pe = cfg->proxy; pe < cfg->proxy + cfg->nproxy; ++pe) {
        pe->s = socket(PF_INET, SOCK_STREAM, 0);
        if (pe->s == -1) {
            log_print(LOG_ERR, "socket: %m; aborting");
            return -1;
        } else if (bind(pe->s, (struct sockaddr*)&pe->from, sizeof pe->from) == -1) {
            log_print(LOG_ERR, "bind(%s:%d): %m; aborting", inet_ntoa(pe->from.sin_addr), (int)ntohs(pe->from.sin_port));
            return -1;
        } else if (listen(pe->s, SOMAXCONN) == -1) {
            log_print(LOG_ERR, "listen(%s:%d): %m; aborting", inet_ntoa(pe->from.sin_addr), (int)ntohs(pe->from.sin_port));
            return -1;
        } else if (fcntl(pe->s, F_SETFL, O_NONBLOCK) == -1) {
            log_print(LOG_ERR, "fcntl(%s:%d): %m; aborting", inet_ntoa(pe->from.sin_addr), (int)ntohs(pe->from.sin_port));
            return -1;
        }
    }
    
    /* Setup the TLS stuff. */
    if (!(ctx = tls_start(cfg)))
        return -1;

    /* Set a nonprivileged UID and GID. */
    if (getuid() == 0) {
        setgid(cfg->gid);
        setuid(cfg->uid);
    }

    if (cfg->pidfile) {
        if (!pidfile_create(cfg)) {
            log_print(LOG_ERR, "can't write PID file; aborting");
            return -1;
        }
    }

    if (do_daemon) {
        log_stderr = 0;
        daemon(0, 1);
    }

    if (cfg->pidfile)
        pidfile_write_pid();

    /* Start up the proxy. */
    restart = do_proxy_accept(cfg, ctx);

    if (cfg->pidfile)
        pidfile_remove(cfg);

    /* Maybe restart on SIGHUP. */
    if (restart) {
        execv(argv[0], argv);
        log_print(LOG_ERR, "could not restart after signal: %m");
        return -1;
    } else
        return 0;
}
